
import asyncio
from typing import Callable

class CallObj:

    def __init__(self, method: Callable, args, kwargs):
        self.method = method
        self.args = args or ()
        self.kwargs = kwargs or {} 

    def add_args(self, *args):
        self.args = (*self.args, *args) # type: ignore
        return self

    def add_kwargs(self, **kwargs):
        self.kwargs = {**self.kwargs, **kwargs} # type: ignore
        return self

    async def __call__(self):
        if asyncio.iscoroutinefunction(self.method):
            return await self.method(*self.args, **self.kwargs)
        return self.method(*self.args, **self.kwargs)


from abc import ABC, abstractmethod

from typing import Optional

from lacia.network.abcbase import BaseServer, BaseClient
from lacia.standard.abcbase import Namespace

class BaseJsonRpc(ABC):
    
    _server: Optional[BaseServer] = None
    _client: Optional[BaseClient] = None
    _Executor: bool = False
    _namespace: Namespace

    @abstractmethod
    def run_server(self, server: BaseServer) -> None:
        ...
    
    @abstractmethod
    def run_client(self, client: BaseClient) -> None:
        ...
    
    @abstractmethod
    def is_server(self) -> bool:
        ...



import asyncio
import time

from lacia.logger import logger
from lacia.standard.jsonast.runtime import RunTime, Standard, Namespace, JsonAst
from lacia.core.proxy import ProxyObj, ResultProxy


async def test_aping(x):
    return f"pong: {x}"

def test_iter(n):
    for i in range(n):
        time.sleep(0.1)
        yield i

async def test_async_iter(n):
    for i in range(n):
        await asyncio.sleep(0.1)
        yield i

async def test_async_iter_send(n):
    j = 0
    for i in range(n):
        await asyncio.sleep(0.1)
        j = yield i + j


class TestJson:
    def __init__(self, a, b=1, c: dict = {}) -> None:
        self.a = a
        self.b = b
        self.c = c

    def sum(self):
        return self.a + self.b

    def test_return(self, text):
        return f"return: {text}"


namespace = {
    "test_attribute": "Hello World!",
    "test_ping": lambda x: f"pong: {x}",
    "test_aping": test_aping,
    "test_class_init": TestJson,
    "test_async_iter": test_async_iter,
    "test_iter": test_iter,
    "test_async_iter_send": test_async_iter_send,
    "Test": TestJson,
}
#  嵌套


class Test:
    async def test_attribute(self):
        runtime = Standard.runtime(namespace, ProxyObj, ResultProxy)
        obj = ProxyObj()

        astboj = obj.test_attribute._obj

        r = await runtime.run(astboj)

        assert r == "Hello World!"

    async def test_ping(self):
        runtime = Standard.runtime(namespace, ProxyObj, ResultProxy)
        obj = ProxyObj()

        astboj = obj.test_ping("hello")._obj

        r = await runtime.run(astboj)

        assert r == "pong: hello"

    async def test_aping(self):
        runtime = Standard.runtime(namespace, ProxyObj, ResultProxy)
        obj = ProxyObj()

        astboj = obj.test_aping("hello")._obj

        r = await runtime.run(astboj)

        assert r == "pong: hello"

    async def test_class_init(self):
        runtime = Standard.runtime(namespace, ProxyObj, ResultProxy)
        obj = ProxyObj()

        astboj = obj.test_class_init(1, b=3).sum()._obj

        r = await runtime.run(astboj)

        assert r == 4

    async def test_class_attribute(self):
        runtime = Standard.runtime(namespace, ProxyObj, ResultProxy)
        obj = ProxyObj()

        astobj = obj.test_class_init(1, b=3).a._obj

        r = await runtime.run(astobj)

        assert r == 1

    async def test_object_nest(self):
        runtime = Standard.runtime(namespace, ProxyObj, ResultProxy)
        a_obj = ProxyObj().test_ping("ping")
        b_obj = ProxyObj().test_class_init(1, b=3).test_return(a_obj)
        c_obj = ProxyObj()

        astobj = c_obj.test_ping(b_obj)._obj

        r = await runtime.run(astobj)

        assert r == "pong: return: pong: ping"

    async def test_getattr(self):
        runtime = Standard.runtime(namespace, ProxyObj, ResultProxy)
        obj = getattr(ProxyObj(), "test_ping")("ping")

        astobj = obj._obj

        r = await runtime.run(astobj)

        assert r == "pong: ping"

    async def test_loads_dumps(self):
        runtime = Standard.runtime(namespace, ProxyObj, ResultProxy)
        obj = ProxyObj()

        astobj = obj.Test(1, b=2, c={"c": 3})._obj

        astobj = JsonAst.loads(JsonAst.dumps(astobj))

        r = await runtime.run(astobj)

        assert r.c == {"c": 3}

    async def test_async_iter(self):
        runtime = Standard.runtime(namespace, ProxyObj, ResultProxy)
        obj = ProxyObj()

        astobj = obj.test_async_iter(3)._obj

        r = await runtime.run(astobj)

        assert [i async for i in r] == [0, 1, 2]

    async def test_iter(self):
        runtime = Standard.runtime(namespace, ProxyObj, ResultProxy)
        obj = ProxyObj()

        astobj = obj.test_iter(3)._obj

        r = await runtime.run(astobj)

        assert [i for i in r] == [0, 1, 2]

    async def test_async_iter_send(self):
        runtime = Standard.runtime(namespace, ProxyObj, ResultProxy)
        obj = ProxyObj()

        astobj = obj.test_async_iter_send(3)._obj

        r = await runtime.run(astobj)

        value = await r.asend(None)
        assert value == 0
        value = await r.asend(value + 1)
        assert value == 2
        value = await r.asend(value + 1)
        assert value == 5



    async def main(self):

        for func in dir(self):
            if func.startswith("test_"):
                await getattr(self, func)()
                logger.success(f"{func} passed")


async def main():

    await Test().main()


asyncio.run(main())




class JsonRpcJsonRpcException(Exception):
    ...

class JsonRpcInitException(JsonRpcJsonRpcException):
    ...

class JsonRpcWsException(JsonRpcJsonRpcException):
    ...

class JsonRpcWsConnectException(JsonRpcWsException):
    ...

class JsonRpcRuntimeException(JsonRpcWsException):
    ...


import asyncio
from typing import Optional

import orjson
import bson
import aiohttp

from lacia.network.abcbase import BaseClient
from lacia.logger import logger
from lacia.types import Message
from lacia.exception import JsonRpcWsConnectException


class AioClient(BaseClient):
    def __init__(
        self,
        path: str = "",
        host: str = "localhost",
        port: int = 8080,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        session: Optional[aiohttp.ClientSession] = None,
        **kwargs,
    ) -> None:
        self.path = path
        self.host = host
        self.port = port
        self.loop = loop
        self.session = session
        self.kwargs = kwargs

    async def start(self) -> "AioClient":
        if self.session is None:
            self.session = aiohttp.ClientSession(
                loop=self.loop or asyncio.get_event_loop(), **self.kwargs
            )
        self.ws = await self.session.ws_connect(
            f"http://{self.host}:{self.port}{self.path}"
        )
        logger.success(
            f"📡 {self.__class__.__name__} success connected: http://{self.host}:{self.port}{self.path}."
        )
        return self

    async def receive(self):
        try:
            async for data in self.ws:
                if data.type == aiohttp.WSMsgType.close:
                    raise JsonRpcWsConnectException(data.data)
                else:
                    return data
        except Exception as e:
            raise JsonRpcWsConnectException(f"{self.__class__.__name__} closed.({e})")

    async def receive_json(self):
        data = await self.receive()
        if data and data.type == aiohttp.WSMsgType.TEXT:
            data = orjson.loads(data.data)
            return data
        elif data and data.type == aiohttp.WSMsgType.BINARY:
            data = bson.loads(data.data)
            return data
        raise JsonRpcWsConnectException("Invalid data type.")

    async def receive_bytes(self):
        data = await self.receive()
        if data and data.type == aiohttp.WSMsgType.BINARY:
            return data.data

    async def iter_json(self):
        try:
            while True:
                data = await self.receive_json()
                if data:
                    yield data
        except JsonRpcWsConnectException as e:
            logger.error(f"http://{self.host}:{self.port}{self.path} closed.")
            await self.close()
        except Exception as e:
            logger.error(e)
            logger.error(f"http://{self.host}:{self.port}{self.path} closed.")
            await self.close()
        finally:
            return

    async def iter_bytes(self):
        try:
            while True:
                data = await self.receive_bytes()
                if data:
                    yield data
        except JsonRpcWsConnectException as e:
            logger.info(f"http://{self.host}:{self.port}{self.path} closed.")
            await self.close()
        except Exception as e:
            logger.error(e)
            logger.info(f"http://{self.host}:{self.port}{self.path} closed.")
            await self.close()

    async def send(self, message) -> None:
        return await self.send(message)

    async def send_bytes(self, message: bytes):
        return await self.ws.send_bytes(message)

    async def send_json(self, message: Message, binary: bool = True):
        if binary:
            return await self.ws.send_bytes(bson.dumps(message))
        return await self.ws.send_json(message)

    async def close(self) -> None:
        await self.ws.close()

    def closed(self) -> bool:
        return self.ws.closed



from typing import Generic, TypeVar, cast, Type, Optional, Any, TYPE_CHECKING

from lacia.standard import JsonAst
from lacia.types import RpcMessage, JsonRpcCode
from lacia.standard.abcbase import BaseDataTrans
from lacia.exception import JsonRpcRuntimeException

if TYPE_CHECKING:
    from lacia.core.core import JsonRpc

T = TypeVar("T", bound=BaseDataTrans)

Schema = TypeVar("Schema")


def with_schema(obj, schema: Type[Schema]) -> Schema:
    return cast(schema, obj)


class BaseProxy(Generic[T]):
    _jsonrpc: str
    _obj: T
    _vision: bool

    def __await__(self):
        ...


class ProxyObj(BaseProxy):
    _jsonrpc: str = "jsonast"

    def __init__(
        self,
        core: Optional["JsonRpc[JsonAst]"] = None,
        name: Optional[str] = None,
        vision: bool = True,
    ):
        self._aiter: Optional["ResultProxy"] = None
        self._obj = ["server", None] if name is None else ["client", name]
        self._core = core
        self._name = name
        self._vision = vision

    def __getattr__(self, name: str) -> "ProxyObj":
        self = self._newobj()

        obj = JsonAst(
            obj=self._obj,
            method="__getattr__",
            args=(name,),
            kwargs={},
        )
        self._obj = obj
        return self

    def __call__(self, *args, **kwargs) -> "ProxyObj":
        self = self._newobj()

        n_args = []
        n_kwargs = {}
        for i in args:
            n_args.append(self._expand(i))

        for k, v in kwargs.items():
            n_kwargs[k] = self._expand(v)

        obj = JsonAst(
            obj=self._obj,
            method="__call__",
            args=tuple(n_args),
            kwargs=n_kwargs,
        )
        self._obj = obj
        return self

    def _expand(self, obj: Any):
        if isinstance(obj, ProxyObj):
            return obj._obj
        elif isinstance(obj, list):
            return [self._expand(i) for i in obj]
        elif isinstance(obj, dict):
            return {k: self._expand(v) for k, v in obj.items()}
        elif isinstance(obj, tuple):
            return tuple(self._expand(i) for i in obj)
        elif isinstance(obj, set):
            return {self._expand(i) for i in obj}
        else:
            return obj

    def _newobj(self):
        new = ProxyObj(self._core, self._name, self._vision)
        setattr(new, "_obj", self._obj)
        setattr(new, "_aiter", self._aiter)
        return new

    def __aiter__(self):
        self = self._newobj()

        obj = JsonAst(obj=self._obj, method="__aiter__", args=(), kwargs={})
        self._obj = obj

        return self

    def __anext_proxy__(self):
        self = self._newobj()

        obj = JsonAst(obj=self._obj, method="__anext__", args=(), kwargs={})
        self._obj = obj

        return self

    async def __anext__(self):
        if self._core is None:
            raise TypeError("ProxyObj is not bind to JsonRpc")
        if not self._aiter:
            if self._core._client is not None:
                data = await self._core.run(self)
            elif self._core._server is not None:
                if self._name is None:
                    raise JsonRpcRuntimeException("client name is None")
                data = await self._core.reverse_run(self._name, self)
            else:
                raise JsonRpcRuntimeException("server and client are None")
            self._aiter = data
            r = await data.__anext__()
            return r
        else:
            r = await self._aiter.__anext__()
            return r

    def __await__(self):
        if self._core is None:
            raise TypeError("ProxyObj is not bind to JsonRpc")
        if self._core._client is not None:
            data = yield from self._core.run(self).__await__()
        elif self._core._server is not None:
            if self._name is None:
                raise JsonRpcRuntimeException("client name is None")
            data = yield from self._core.reverse_run(self._name, self).__await__()
        else:
            raise JsonRpcRuntimeException("server and client are None")
        self._obj = None
        if self._vision:
            return data.visions
        return data


class ResultProxy(BaseProxy):
    def __init__(
        self,
        result: RpcMessage,
        core: Optional["JsonRpc[JsonAst]"] = None,
        by: Optional[str] = None,
        vision: bool = True,
    ):
        self._core = core
        self._result = result
        self._by = by
        self._obj = None
        self._vision = vision

    @property
    def visions(self):
        return self._raise_result()

    def _raise_result(self, vision: bool = True):
        if self._result.error:
            if self._result.error_code == JsonRpcCode.StopAsyncIterationError:
                raise StopAsyncIteration
            raise JsonRpcRuntimeException(self._result.error)
        return self._result.result if vision else self

    def __getattr__(self, name: str) -> "ProxyObj":
        return getattr(
            getattr(ProxyObj(self._core, self._by, self._vision), self._result.id), name
        )

    async def __aiter__(self):
        return self

    async def __anext__(self):
        if self._core is None:
            raise TypeError("ProxyObj is not bind to JsonRpc")
        obj = getattr(ProxyObj(self._core, self._by), self._result.id).__anext_proxy__()

        if self._core._client is not None:
            data = await self._core.run(obj)
        elif self._core._server is not None:
            if self._by is None:
                raise JsonRpcRuntimeException("client name is None")
            data = await self._core.reverse_run(self._by, obj)
        else:
            raise JsonRpcRuntimeException("server and client are None")
        return data._raise_result(self._vision)


def set_vision(result: ResultProxy, proxy: BaseProxy[BaseDataTrans]):
    result._vision = proxy._vision
    return result



from typing import TypeVar, Generic, Type, Any, Dict, NamedTuple, Hashable, Protocol
from typing_extensions import Self

T = TypeVar("T")
S = TypeVar("S", contravariant=True)

class Namespace(NamedTuple):

    builtins: Dict[str, Any] = {}
    globals: Dict[str, Any] = {}
    locals: Dict[Hashable, Dict[str, Any]] = {}

    def __getitem__(self, key: T) -> Dict[str, Any]:

        return {
            **self.builtins,
            **self.globals,
            **self.locals.get(key, {})
        }      

class BaseDataTrans(Protocol[T]):

    @classmethod
    def loads(cls, obj: T) -> Self:
        ...
    
    def dumps(self) -> T:
        ...

class BaseRunTime(Generic[S]):

    def __init__(self, namespace: Dict[str, Any], proxy, proxyresult):
        self.namespace = namespace
        self.proxy = proxy
        self.proxyresult = proxyresult

    async def run(self, obj: S) -> Any:
        ...

class BaseStandard(Protocol[T, S]):
    
    datatrans: Type[BaseDataTrans[T]]
    runtime: Type[BaseRunTime[S]]



import asyncio
from typing import Optional, Dict, TYPE_CHECKING

import orjson
import bson
import aiohttp
from aiohttp import web, WSCloseCode

from lacia.network.abcbase import BaseServer, Connection
from lacia.logger import logger
from lacia.utils.tool import CallObj
from lacia.exception import JsonRpcWsConnectException

if TYPE_CHECKING:
    from lacia.core.core import JsonRpc


class AioServer(BaseServer[web.WebSocketResponse]):
    on_events = {}

    def __init__(
        self,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ) -> None:
        self.active_connections: Connection[web.WebSocketResponse] = Connection()
        self.name_connections: Dict[str, web.WebSocketResponse] = {}
        self.loop = loop

    async def websocket_handler(self, request):
        event = asyncio.Event()
        ws = web.WebSocketResponse(autoclose=False)
        await ws.prepare(request)
        self.active_connections.set_ws(ws, event)

        logger.success(f"{str(ws)} connected.")

        headers = request.headers

        obj = self.on_events.get("connect")
        if obj is not None:
            await obj.method(ws, dict(headers), *obj.args, **obj.kwargs)

        await event.wait()
        return ws

    def disconnect(self, websocket: web.WebSocketResponse):
        for ws, event in self.active_connections.ws.items():
            if ws == websocket:
                event.set()
                self.active_connections.clear_ws(ws)
                break

    async def receive(self, websocket: web.WebSocketResponse):
        try:
            async for data in websocket:
                if data.type == aiohttp.WSMsgType.close:
                    raise JsonRpcWsConnectException(str(data.data))
                else:
                    return data
        except Exception as e:
            await self.close_ws(websocket)
            raise JsonRpcWsConnectException(f"{self.__class__.__name__} closed.")

    async def receive_json(self, websocket: web.WebSocketResponse):
        data = await self.receive(websocket)
        if data and data.type == aiohttp.WSMsgType.TEXT:
            data = orjson.loads(data.data)
            return data
        elif data and data.type == aiohttp.WSMsgType.BINARY:
            data = bson.loads(data.data)
            return data
        raise JsonRpcWsConnectException("Invalid data type.")

    async def receive_bytes(self, websocket: web.WebSocketResponse):
        data = await self.receive(websocket)
        if data and data.type == aiohttp.WSMsgType.BINARY:
            return data.data

    async def iter_bytes(self, websocket: web.WebSocketResponse):
        ws_name = str(websocket)
        try:
            while True:
                data = await self.receive_bytes(websocket)
                if data:
                    yield data
        except JsonRpcWsConnectException:
            logger.info(f"{ws_name} disconnected.")
        except Exception as e:
            print(e)
            logger.info(f"{ws_name} disconnected.")
        finally:
            return

    async def iter_json(self, websocket: web.WebSocketResponse):
        ws_name = str(websocket)
        try:
            while True:
                data = await self.receive_json(websocket)
                if data:
                    yield data
        except JsonRpcWsConnectException:
            logger.info(f"{ws_name} disconnected.")
        except Exception as e:
            print(e)
            logger.info(f"{ws_name} disconnected.")
        finally:
            return

    async def send_json(
        self, websocket: web.WebSocketResponse, message: dict, binary: bool = True
    ):
        if binary:
            return await websocket.send_bytes(bson.dumps(message))
        return await websocket.send_json(message)

    async def send_bytes(self, websocket: web.WebSocketResponse, message: bytes):
        return await websocket.send_bytes(message)

    async def close_ws(self, websocket: web.WebSocketResponse):
        name = str(websocket)
        obj = self.on_events.get("disconnect")
        if obj is not None:
            await obj.method(websocket, *obj.args, **obj.kwargs)
        self.disconnect(websocket)
        logger.info(f"{name} disconnected.")

    async def on_shutdown(self):
        for ws, event in self.active_connections.ws.items():
            await ws.close(code=WSCloseCode.GOING_AWAY, message=b"Server shutdown")
            event.set()

    def on(
        self,
        event: str,
        func,
        args: Optional[tuple] = None,
        kwargs: Optional[dict] = None,
    ) -> None:
        self.on_events[event] = CallObj(method=func, args=args, kwargs=kwargs)


def mount_app(
    app: web.Application,
    server: AioServer,
    rpc: "JsonRpc",
    path: str = "/ws",
    **kwargs,
):
    async def startup_func(app):
        await rpc.run_server(server)

    app.on_startup.append(startup_func)
    app.add_routes([web.get(path, server.websocket_handler, **kwargs)])



import asyncio
from typing import Optional, Dict, TYPE_CHECKING

import orjson
import bson
from starlette.websockets import WebSocket

from lacia.network.abcbase import BaseServer, Connection
from lacia.logger import logger
from lacia.utils.tool import CallObj
from lacia.exception import JsonRpcWsConnectException

if TYPE_CHECKING:
    from lacia.core.core import JsonRpc

    from starlette.applications import Starlette
    from starlette.routing import WebSocketRoute


class StarletteServer(BaseServer[WebSocket]):
    on_events = {}

    def __init__(
        self,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ) -> None:
        self.active_connections: Connection[WebSocket] = Connection()
        self.name_connections: Dict[str, WebSocket] = {}
        self.loop = loop

    async def websocket_handler(self, websocket: WebSocket):
        event = asyncio.Event()
        await websocket.accept()
        self.active_connections.set_ws(websocket, event)

        logger.success(f"{str(websocket)} connected.")

        headers = websocket.headers

        obj = self.on_events.get("connect")
        if obj is not None and obj.method is not None:
            await obj.method(websocket, dict(headers), *obj.args, **obj.kwargs)

        await event.wait()

    def disconnect(self, websocket: WebSocket):
        for ws, event in self.active_connections.ws.items():
            if ws == websocket:
                event.set()
                self.active_connections.clear_ws(ws)
                break

    async def receive(self, websocket: WebSocket):
        try:
            data = await websocket.receive()
            if data["type"] == "websocket.disconnect":
                raise JsonRpcWsConnectException(str(data["code"]))
            else:
                return data
        except Exception as e:
            await self.close_ws(websocket)
            raise JsonRpcWsConnectException(f"{self.__class__.__name__} closed.")

    async def receive_json(self, websocket: WebSocket):
        data = await self.receive(websocket)
        if data["type"] == "websocket.receive" and "text" in data:
            return orjson.loads(data["text"])
        elif data["type"] == "websocket.receive" and "bytes" in data:
            return bson.loads(data["bytes"])
        raise JsonRpcWsConnectException("Invalid data type.")

    async def receive_bytes(self, websocket: WebSocket):
        data = await self.receive(websocket)
        if data["type"] == "websocket.receive" and "bytes" in data:
            return data["bytes"]

    async def iter_bytes(self, websocket: WebSocket):
        ws_name = str(websocket)
        try:
            while True:
                data = await self.receive_bytes(websocket)
                if data:
                    yield data
        except JsonRpcWsConnectException:
            logger.info(f"{ws_name} disconnected.")
        except Exception as e:
            print(e)
            logger.info(f"{ws_name} disconnected.")
        finally:
            return

    async def iter_json(self, websocket: WebSocket):
        ws_name = str(websocket)
        try:
            while True:
                data = await self.receive_json(websocket)
                if data:
                    yield data
        except JsonRpcWsConnectException:
            logger.info(f"{ws_name} disconnected.")
        except Exception as e:
            print(e)
            logger.info(f"{ws_name} disconnected.")
        finally:
            return

    async def send_json(self, websocket: WebSocket, message: dict, binary: bool = True):
        if binary:
            await websocket.send_bytes(bson.dumps(message))
        else:
            await websocket.send_text(orjson.dumps(message).decode())

    async def send_bytes(self, websocket: WebSocket, message: bytes):
        await websocket.send_bytes(message)

    async def close_ws(self, websocket: WebSocket):
        name = str(websocket)
        obj = self.on_events.get("disconnect")
        if obj is not None and obj.method is not None:
            await obj.method(websocket, *obj.args, **obj.kwargs)
        self.disconnect(websocket)
        await self.on_shutdown()
        logger.info(f"{name} disconnected.")

    async def on_shutdown(self):
        for ws, event in self.active_connections.ws.items():
            await ws.close(code=1001)
            event.set()

    def on(
        self,
        event: str,
        func,
        args: Optional[tuple] = None,
        kwargs: Optional[dict] = None,
    ) -> None:
        self.on_events[event] = CallObj(method=func, args=args, kwargs=kwargs)


def mount_app(
    app: "Starlette",
    server: StarletteServer,
    rpc: "JsonRpc",
    path: str = "/ws",
    **kwargs,
):
    async def startup_func():
        await rpc.run_server(server)

    app.add_event_handler("startup", startup_func)
    app.add_websocket_route(path, server.websocket_handler, **kwargs)



import asyncio
import json
import richuru
from uuid import uuid4
from typing import Dict, Any, Optional, TypeVar, Generic

from nest_asyncio import apply as nest_apply

from lacia.core.abcbase import BaseJsonRpc
from lacia.core.proxy import BaseProxy, ResultProxy, ProxyObj, set_vision
from lacia.network.abcbase import BaseServer, BaseClient
from lacia.standard.abcbase import BaseDataTrans, Namespace
from lacia.standard.execute import Standard
from lacia.logger import logger
from lacia.types import RpcMessage, Context
from lacia.exception import JsonRpcInitException

T = TypeVar("T")


class JsonRpc(BaseJsonRpc, Generic[T]):
    def __init__(
        self,
        name: str,
        execer: bool = True,
        namespace: Optional[Dict[str, Any]] = None,
        token: Optional[str] = None,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        debug: bool = False,
    ) -> None:
        self._name = name
        self._execer = execer
        self._namespace = Namespace(
            globals=namespace if namespace else {},
        )
        self._token = token
        self._loop = loop
        self._debug = debug
        self._uuid = str(uuid4())

        self._wait_remote: Dict[str, asyncio.Event] = {}
        self._wait_result: Dict[str, ResultProxy] = {}

        self._standard = Standard()

        if self._debug:
            richuru.install(level="DEBUG")

    def add_namespace(self, namespace: Dict[str, Any]) -> None:
        self._namespace.globals.update(namespace)

    async def run_client(self, client: BaseClient) -> None:
        self._standard.init_standard()
        self._client = client
        self._loop = self._loop or asyncio.get_event_loop()
        await client.start()
        if self._loop:
            self._loop.create_task(self._listening_server(self._client.ws))
            logger.info("run client")

    async def run_server(self, server: BaseServer) -> None:
        self._loop = self._loop or asyncio.get_event_loop()
        # nest_apply()
        self._standard.init_standard()
        self._server = server
        self._server.on("connect", self._listening_client)
        self._server.on("disconnect", self.on_server_close)
        logger.info("run server")

    async def _listening_client(self, websocket: T, headers: dict[str, Any]):
        logger.info("listening client")
        Context.websocket.set(websocket)
        Context.namespace.set(self._namespace)
        Context.rpc.set(self)
        Context.headers.set(headers)
        by_name: Optional[str] = None

        event = asyncio.Event()

        def rpc_auto_register(name, token):
            nonlocal by_name
            if self._server is not None:
                self._server.active_connections.set_name_ws(name, websocket)
            Context.name.set(name)
            by_name = name
            if token == self._token:
                event.set()
                return self._name
            raise JsonRpcInitException("rpc_auto_register fail")

        if self._namespace.locals.get(websocket) is None:
            self._namespace.locals[websocket] = {}
        if self._server:
            self._namespace.locals[websocket]["rpc_auto_register"] = rpc_auto_register

        qmgs = asyncio.Queue()

        if self._loop:
            self._loop.create_task(self._client_auth(event, qmgs, websocket))

        if self._server is not None:
            async for message in self._server.iter_json(websocket):
                logger.debug(f"receive: {message}")
                msg = RpcMessage(message)
                if msg.is_request and self._execer and self._loop:
                    if msg.is_auth:
                        self._loop.create_task(self._s_execute(websocket, msg))
                    else:
                        qmgs.put_nowait(msg)
                elif msg.is_response:
                    rmsg = ResultProxy(msg, core=self, by=by_name)  # type: ignore
                    self._wait_result[msg.id] = rmsg
                    self._wait_remote[msg.id].set()
        else:
            raise JsonRpcInitException("server is None")

    async def _listening_server(self, websocket: T):
        logger.info("listening server")
        Context.websocket.set(websocket)
        Context.namespace.set(self._namespace)
        Context.rpc.set(self)
        Context.name.set(None)  # type: ignore

        if self._client is not None:
            if self._loop:
                self._loop.create_task(
                    self.run(ProxyObj().rpc_auto_register(self._name, self._token))
                )

            async for message in self._client.iter_json():
                logger.debug(f"receive: {message}")
                msg = RpcMessage(message)

                if msg.is_request and self._execer and self._loop:
                    self._loop.create_task(self._c_execute(websocket, msg))
                elif msg.is_response:
                    rmsg = ResultProxy(msg, core=self, by=None)  # type: ignore
                    self._wait_result[msg.id] = rmsg
                    self._wait_remote[msg.id].set()
        else:
            raise JsonRpcInitException("client is None")

    async def _s_execute(self, websocket: T, message: RpcMessage):
        result, error = await self._standard.rpc_request(
            message.data, self._namespace[websocket], ProxyObj, ResultProxy
        )

        if error is None:
            if websocket not in self._namespace.locals:
                self._namespace.locals[websocket] = {}
            self._namespace.locals[websocket][message.id] = result
            msg = {
                "jsonrpc": message.jsonrpc,
                "id": message.id,
                "result": self._pretreatment(result),
            }
        else:
            msg = {"jsonrpc": message.jsonrpc, "id": message.id, "error": error}
        if self._server is not None:
            logger.debug(f"send: {msg}")

            await self._server.send_json(websocket, msg)

    async def _c_execute(self, websocket: T, message: RpcMessage):
        result, error = await self._standard.rpc_request(
            message.data, self._namespace[websocket], ProxyObj, ResultProxy
        )

        if error is None:
            if websocket not in self._namespace.locals:
                self._namespace.locals[websocket] = {}
            self._namespace.locals[websocket][message.id] = result  # TODO 内存泄漏
            msg = {
                "jsonrpc": message.jsonrpc,
                "id": message.id,
                "result": self._pretreatment(result),
            }
        else:
            msg = {"jsonrpc": message.jsonrpc, "id": message.id, "error": error}

        if self._client is not None:
            logger.debug(f"send: {msg}")
            await self._client.send_json(msg)

    def on_client_close(self, websocket: T):
        del self._namespace.locals[websocket]

    def on_server_close(self, websocket: T):
        del self._namespace.locals[websocket]

    async def run(self, proxy: BaseProxy[BaseDataTrans]):
        uuid_str = str(uuid4())
        if proxy._obj is None:
            raise JsonRpcInitException("proxy._obj is None")
        data = proxy._obj.dumps()
        event = asyncio.Event()

        self._wait_remote[uuid_str] = event

        msg = {
            "jsonrpc": proxy._jsonrpc,
            "id": uuid_str,
            "method": data,
        }

        if self._client is not None:
            logger.debug(f"send: {msg}")
            await self._client.send_json(msg)
        else:
            raise JsonRpcInitException("server and client are None R")

        await event.wait()

        res = self._wait_result.pop(uuid_str)
        set_vision(res, proxy)
        return res

    async def reverse_run(self, name: str, proxy: BaseProxy[BaseDataTrans]):
        uuid_str = str(uuid4())
        if proxy._obj is None:
            raise JsonRpcInitException("proxy._obj is None")
        data = proxy._obj.dumps()
        event = asyncio.Event()

        self._wait_remote[uuid_str] = event

        msg = {
            "jsonrpc": proxy._jsonrpc,
            "id": uuid_str,
            "method": data,
        }

        if self._server is not None:
            logger.debug(f"send: {msg}")
            await self._server.send_json(
                self._server.active_connections.get_ws(name), msg
            )
        else:
            raise JsonRpcInitException("server and client are None S")

        await event.wait()

        res = self._wait_result.pop(uuid_str)
        set_vision(res, proxy)
        return res

    async def _client_auth(
        self, event: asyncio.Event, qmgs: asyncio.Queue, websocket: T
    ):
        await event.wait()
        if self._server is not None:
            Context.name.set(self._server.active_connections.get_name(websocket))
        while True:
            msg = await qmgs.get()
            await self._s_execute(websocket, msg)

    def _pretreatment(self, data: Any) -> Any:
        class BytesEncoder(json.JSONEncoder):
            def default(self, obj):
                if isinstance(obj, bytes):
                    return None
                return super().default(obj)

        try:
            json.dumps(data, cls=BytesEncoder)
            return data
        except TypeError:
            return str(data)

    def is_server(self) -> bool:
        if self._server is None and self._client is None:
            raise JsonRpcInitException("server and client are None")
        elif self._client is None and self._server is not None:
            return True
        return False

    @property
    def jsonast(self):
        return ProxyObj(self)  # type: ignore



import asyncio
from typing import Dict, Any, Type, TYPE_CHECKING

from lacia.standard.abcbase import BaseRunTime, BaseStandard, Namespace
from lacia.standard.jsonast.impl import JsonAst
from lacia.types import Context

if TYPE_CHECKING:
    from lacia.core.proxy import ProxyObj, ResultProxy


class RunTime(BaseRunTime[JsonAst]):
    def __init__(
        self,
        namespace: Dict[str, Any],
        proxy: Type["ProxyObj"],
        proxyresult: Type["ResultProxy"],
    ):
        self.namespace = namespace
        self.proxy = proxy
        self.proxyresult = proxyresult
        try:
            self.rpc = Context.rpc.get()
            self.is_server = self.rpc.is_server()
            self.self_name = self.rpc._name
        except Exception:
            self.rpc = None
            self.is_server = True
            self.self_name = None

    async def run(self, ast):
        if isinstance(ast, JsonAst):
            if isinstance(ast.obj, JsonAst):
                obj = await self.run(ast.obj)
            elif (
                ast.obj is None
                or (ast.obj == ["server", None] and self.is_server)
                or (ast.obj == ["client", self.self_name] and not self.is_server)
            ):
                if ast.args is not None:
                    if len(ast.args) == 1 and isinstance(ast.args[0], str):
                        obj = self.namespace[ast.args[0]]
                        return obj
                    else:
                        raise TypeError(f"obj type error: {ast.obj}")
                else:
                    raise TypeError(f"obj type error: {ast.obj}")
            elif isinstance(ast.obj, str):
                obj = self.namespace[ast.obj]
            elif (
                isinstance(ast.obj, list)
                and len(ast.obj) == 2
                and ast.obj[0] == "server"
            ):
                obj = self.proxy(self.rpc, vision=False)
            elif (
                isinstance(ast.obj, list)
                and len(ast.obj) == 2
                and ast.obj[0] == "client"
            ):
                obj = self.proxy(self.rpc, ast.obj[1], vision=False)
            else:
                raise TypeError(f"obj type error: {ast.obj}")
            if ast.method is None or ast.obj is None:
                return obj
            if ast.args is not None:
                if (
                    len(ast.args) == 1
                    and ast.method == "__getattr__"
                    and isinstance(ast.args[0], str)
                ):
                    return getattr(obj, ast.args[0])
            func = getattr(obj, ast.method)
            if ast.args is None and ast.kwargs is None:
                return func
            if ast.method == "__anext__":
                return await func()

            args = []
            kwargs = {}
            for i in ast.args or ():
                args.append(await self.run(i))
            if ast.kwargs:
                for k, v in ast.kwargs.items():
                    kwargs[k] = await self.run(v)
            args = tuple(args)

            if ast.method == "__call__" and asyncio.iscoroutinefunction(obj):
                return await func(*args, **kwargs)
            elif asyncio.iscoroutinefunction(func):
                return await func(*args, **kwargs)
            elif isinstance(obj, self.proxy):
                return func(*args, **kwargs)
            return func(*args, **kwargs)
        elif isinstance(ast, self.proxy):
            return await ast
        elif isinstance(ast, self.proxyresult):
            return ast.visions
        elif isinstance(ast, (list, tuple)):
            t_args = []
            for i in ast:
                t_args.append(await self.run(i))
            return type(ast)(t_args)
        elif isinstance(ast, dict):
            return {k: await self.run(v) for k, v in ast.items()}
        else:
            return ast


class Standard(BaseStandard[dict, JsonAst]):
    _jsonrpc = "jsonast"

    datatrans: Type[JsonAst] = JsonAst
    runtime: Type[RunTime] = RunTime




from typing import TypeVar, Generic, Dict, ClassVar

from lacia.logger import logger
from lacia.types import JsonRpcCode
from lacia.standard.abcbase import BaseStandard, Namespace, BaseRunTime

S = TypeVar("S")

class Standard(Generic[S]):

    execute: Dict[str, BaseStandard[dict, S]] = {}
    
    @classmethod
    def init_standard(cls):
        for name, obj in cls.get_rpc_subclasses(BaseStandard):
            cls.execute[name] = obj
        
    @classmethod
    def get_all_subclasses(cls, obj):

        for subclass in obj.__subclasses__():
            yield from cls.get_all_subclasses(subclass)
            yield subclass
    
    @classmethod
    def get_rpc_subclasses(cls, obj):

        for subclass in cls.get_all_subclasses(obj):
            if hasattr(subclass, "_jsonrpc"):
                yield getattr(subclass, "_jsonrpc"), subclass 

    @classmethod
    async def rpc_request(cls, data: dict, namespace: dict, proxy, proxyresult):
        jsonrpc = data.get("jsonrpc")
        result, error = None, None
        if jsonrpc is None:
            return None, {"code": JsonRpcCode.ParseError, "message": "jsonrpc is None"}
        elif jsonrpc in cls.execute:
            runtime: BaseRunTime = cls.execute[jsonrpc].runtime(namespace, proxy, proxyresult)
            obj = cls.execute[jsonrpc].datatrans.loads(data["method"]) # type: ignore
            try:
                result = await runtime.run(obj)
                if isinstance(result, proxy):
                    result = await result
                if isinstance(result, proxyresult):
                    if data["method"]["method"] not in ("__aiter__", "__anext__"):
                        result = result.visions
            except StopAsyncIteration as e:
                return result, {"code": JsonRpcCode.StopAsyncIterationError, "message": ""}
            except Exception as e:
                error = e
                logger.error(e)
            return result, {"code": JsonRpcCode.InternalError, "message": str(error)} if error else None
        return None, {"code": JsonRpcCode.MethodNotFound, "message": f"jsonrpc {jsonrpc} not found"}




from __future__ import annotations

from enum import Enum
from contextvars import ContextVar
from typing import Dict, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from lacia.standard.abcbase import Namespace

Message = Dict[str, Any]

class JsonRpcCode(int, Enum):
    ParseError = -32700             
    InvalidRequest = -32600
    MethodNotFound = -32601
    InvalidParams = -32602
    InternalError = -32603
    ServerError = -32000
    StopAsyncIterationError = -32099

class RpcMessage:

    def __init__(self, data: Dict[str, Any]):
        self.data = data
    
    @property
    def jsonrpc(self) -> str:
        return self.data.get("jsonrpc", None)
    
    @property
    def id(self) -> Any:
        return self.data.get("id", None)

    @property
    def method(self):
        return self.data.get("method", {})

    @property
    def result(self) -> Any:
        return self.data.get("result", None)
    
    @property
    def error(self) -> Any:
        return self.data.get("error", None)
    
    @property
    def error_code(self) -> int:
        return self.data["error"]["code"]

    @property
    def error_msg(self) -> str:
        return self.data["error"]["message"]
    
    @property
    def is_error(self) -> bool:
        return "error" in self.data

    @property
    def is_request(self) -> bool:
        return "method" in self.data
    
    @property
    def is_response(self) -> bool:
        return "result" in self.data or "error" in self.data

    @property
    def is_auth(self) -> bool:

        return all(
            [
                self.method["obj"] == {'obj': ["server", None], 'method': '__getattr__', 'args': ['rpc_auto_register'], 'kwargs': {}},
            ]
        )

class Context:
    websocket: ContextVar = ContextVar("websocket")
    name: ContextVar[str] = ContextVar("name")
    namespace: ContextVar[Namespace] = ContextVar("namespace")
    rpc: ContextVar = ContextVar("rpc")
    headers: ContextVar[dict[str, Any]] = ContextVar("headers")

    @classmethod
    def get_cookie(cls) -> dict[str, Any]:
        headers = cls.headers.get()
        if headers is None:
            return {}
        cookie = headers.get("Cookie", {})
        return {
            k: v for k, v in (item.split("=", 1) for item in cookie.split("; "))
        }



import asyncio
from abc import abstractmethod
from typing import Optional, TypeVar, Generic, Generator, Dict, Callable

from lacia.types import Message
from lacia.utils.tool import CallObj

T = TypeVar("T")


class Connection(Generic[T]):
    def __init__(self):
        self.ws: Dict[T, asyncio.Event] = {}
        self.name_ws: Dict[str, T] = {}

    def set_ws(self, ws: T, event: asyncio.Event):
        self.ws[ws] = event

    def set_name_ws(self, name: str, ws: T):
        self.name_ws[name] = ws

    def clear_ws(self, ws: T):
        self.ws.pop(ws)
        for name, _ws in self.name_ws.items():
            if _ws == ws:
                self.name_ws.pop(name)

    def clear_name_ws(self, name: str):
        self.name_ws.pop(name)

    def get_ws(self, name: str) -> T:
        return self.name_ws[name]

    def get_name(self, ws: T) -> str:
        for name, _ws in self.name_ws.items():
            if _ws == ws:
                return name
        raise KeyError("no such websocket")


class BaseServer(Generic[T]):
    active_connections: Connection[T]
    name_connections: Dict[str, T]
    on_events: Dict[str, CallObj]

    @abstractmethod
    async def receive(self, websocket: T) -> Message:
        ...

    async def send(self, message) -> None:
        ...

    @abstractmethod
    async def receive_json(self, websocket: T) -> Message:
        ...

    @abstractmethod
    async def receive_bytes(self, websocket: T) -> bytes:
        ...

    @abstractmethod
    async def iter_bytes(self, websocket: T) -> Generator[bytes, None, None]:
        ...

    @abstractmethod
    async def iter_json(self, websocket: T) -> Generator[Message, None, None]:
        ...

    @abstractmethod
    async def send_bytes(self, websocket: T, message: bytes) -> None:
        ...

    @abstractmethod
    async def send_json(self, websocket: T, message: Message, binary=True) -> None:
        ...

    @abstractmethod
    async def start(self) -> "BaseServer":
        ...

    @abstractmethod
    def on(
        self,
        event: str,
        func: Callable,
        args: Optional[tuple] = None,
        kwargs: Optional[dict] = None,
    ) -> None:
        ...


class BaseClient(Generic[T]):
    ws: T

    @abstractmethod
    async def receive(self) -> Message:
        ...

    @abstractmethod
    async def send(self, message) -> None:
        ...

    @abstractmethod
    async def receive_json(self) -> Message:
        ...

    @abstractmethod
    async def receive_bytes(self) -> bytes:
        ...

    @abstractmethod
    async def iter_bytes(self) -> Generator[bytes, None, None]:
        ...

    @abstractmethod
    async def iter_json(self) -> Generator[Message, None, None]:
        ...

    @abstractmethod
    async def send_bytes(self, message: bytes) -> None:
        ...

    @abstractmethod
    async def send_json(self, message: Message, binary=True) -> None:
        ...

    @abstractmethod
    async def start(self, *args, **kwargs) -> "BaseClient":
        ...

    @abstractmethod
    async def close(self, code: int, reason: Optional[str] = None) -> None:
        ...

    @abstractmethod
    def closed(self) -> bool:
        ...



import inspect
import asyncio

from functools import wraps

from lacia.logger import logger

from typing import (
    Callable,
    TypeVar,
    Dict,
    Any,
    Optional,
    ParamSpec
)


P = ParamSpec("P")
T = TypeVar("T")

class BaseClass:
    __on_obj: Dict[str, Callable] = {}

    @classmethod
    def on(cls, name: str, func: Optional[Callable] = None):
        def decorator(func: Callable[P, T]) -> Callable[P, T]:
            @wraps(func)
            def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                return func(*args, **kwargs)

            @wraps(func)
            async def async_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                return await func(*args, **kwargs)  # type: ignore

            cls.__on_obj[name] = async_wrapper if asyncio.iscoroutinefunction(func) else wrapper  # type: ignore

            return cls.__on_obj[name]  # type: ignore
        if func:
            return decorator(func)
        return decorator

    @classmethod
    def __run(
        cls,
        obj: Any,
        objer: Callable[P, T],
        name: str,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> T:
        if func := cls.__on_obj.get(name, None):
            return func(obj, objer, name, *args, **kwargs)
        return objer(*args, **kwargs)

    @classmethod
    async def __async_run(
        cls,
        obj: Any,
        objer: Callable[P, T],
        name: str,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> T:
        if func := cls.__on_obj.get(name, None):
            return await func(obj, objer, name, *args, **kwargs)
        return await objer(*args, **kwargs)  # type: ignore

    @staticmethod
    def __get_func_bind(func: Callable, *args, **kwargs) -> Dict[str, Any]:
        """
        Get the function binding.
        """
        sig = inspect.signature(func)
        return {k: v for k, v in sig.bind(*args, **kwargs).arguments.items()}

    @classmethod
    def __hook_func(cls, name: str, obj):
        """
        Hook function for the tool.
        """

        def decorator(func: Callable[P, T]) -> Callable[P, T]:
            @wraps(func)
            def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                return cls.__run(obj, func, name, *args, **kwargs)

            @wraps(func)
            async def async_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                return await cls.__async_run(obj, func, name, *args, **kwargs)

            return async_wrapper if asyncio.iscoroutinefunction(func) else wrapper  # type: ignore

        return decorator

    def __getattribute__(self, __name: str):
        if __name.startswith('_'):
            return super().__getattribute__(__name)
        obj = super().__getattribute__(__name)
        if __name in self.__on_obj:
            if callable(obj):
                return self.__hook_func(__name, obj)(obj)
            return self.__run(self, obj, __name)
        return obj


class MetaHook(BaseClass, type):
    ...


class Hook(metaclass = MetaHook):
    __on_obj: Dict[str, Callable] = {}

    @classmethod
    def on(cls, name: str, func: Optional[Callable] = None):
        def decorator(func: Callable[P, T]) -> Callable[P, T]:
            @wraps(func)
            def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                return func(*args, **kwargs)

            @wraps(func)
            async def async_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                return await func(*args, **kwargs)  # type: ignore

            cls.__on_obj[name] = async_wrapper if asyncio.iscoroutinefunction(func) else wrapper  # type: ignore

            return cls.__on_obj[name]  # type: ignore
        if func:
            return decorator(func)
        return decorator

    @classmethod
    def __run(
        cls,
        obj: Any,
        objer: Callable[P, T],
        name: str,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> T:
        if func := cls.__on_obj.get(name, None):
            return func(obj, objer, name, *args, **kwargs)
        return objer(*args, **kwargs)

    @classmethod
    async def __async_run(
        cls,
        obj: Any,
        objer: Callable[P, T],
        name: str,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> T:
        if func := cls.__on_obj.get(name, None):
            return await func(obj, objer, name, *args, **kwargs)
        return await objer(*args, **kwargs)  # type: ignore

    @staticmethod
    def __get_func_bind(func: Callable, *args, **kwargs) -> Dict[str, Any]:
        """
        Get the function binding.
        """
        sig = inspect.signature(func)
        return {k: v for k, v in sig.bind(*args, **kwargs).arguments.items()}

    @classmethod
    def __hook_func(cls, name: str, obj):
        """
        Hook function for the tool.
        """

        def decorator(func: Callable[P, T]) -> Callable[P, T]:
            @wraps(func)
            def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                return cls.__run(obj, func, name, *args, **kwargs)

            @wraps(func)
            async def async_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                return await cls.__async_run(obj, func, name, *args, **kwargs)

            return async_wrapper if asyncio.iscoroutinefunction(func) else wrapper  # type: ignore

        return decorator

    def __getattribute__(self, __name: str):
        if __name.startswith('_'):
            return super().__getattribute__(__name)
        obj = super().__getattribute__(__name)
        if __name in self.__on_obj:
            if callable(obj):
                return self.__hook_func(__name, obj)(obj)
            return self.__run(self, obj, __name)
        return obj


async def receive_json_hook(obj, objer, name, *args, **kwargs):
    res = await objer(*args, **kwargs)
    logger.info(f"{obj.__self__.__class__.__name__} received: {res}")
    return res

async def send_json_hook(obj, objer, name, *args, **kwargs):
    res = await objer(*args, **kwargs)
    logger.info(f"{obj.__self__.__class__.__name__} send: {args[-1] if args else None}")
    return res

# Hook.on('receive_json', receive_json_hook)
# Hook.on('send_json', send_json_hook) 


from typing import NamedTuple, Tuple, Dict, Optional, Any, Union, TypeVar, List

from lacia.standard.abcbase import BaseDataTrans

T = TypeVar("T")

class BaseJsonAst(NamedTuple):
    obj: Union[str, "JsonAst", None, List[str]]
    method: Optional[str]
    args: Optional[Tuple[Any, ...]] # Optional[Tuple[Union[str, "JsonAst"], ...]]
    kwargs: Optional[Dict[str, Any]] # Optional[Dict[str, "JsonAst" | Any]]
    
    def todict(self):
        def handle_value(value):
            if isinstance(value, BaseJsonAst):
                return value.todict()
            elif isinstance(value, (list, tuple)):
                return type(value)(handle_value(v) for v in value)
            elif isinstance(value, dict):
                return {k: handle_value(v) for k, v in value.items()}
            else:
                return value

        return {
            "obj": handle_value(self.obj),
            "method": self.method,
            "args": handle_value(self.args),
            "kwargs": handle_value(self.kwargs),
        }

    @classmethod
    def fromdict(cls, d: dict):
        def handle_value(value):
            if isinstance(value, dict) and cls.is_jsonast(value):
                return cls.fromdict(value)
            elif isinstance(value, (list, tuple)):
                return type(value)(handle_value(v) for v in value)
            elif isinstance(value, dict):
                return {k: handle_value(v) for k, v in value.items()}
            else:
                return value

        return cls(
            obj=handle_value(d["obj"]), # type: ignore
            method=d["method"],
            args=handle_value(d["args"]),  # type: ignore
            kwargs=handle_value(d["kwargs"]),  # type: ignore
        )
    
    @classmethod
    def is_jsonast(cls, obj: Any) -> bool:
        return isinstance(obj, dict) and "obj" in obj and "method" in obj and "args" in obj and "kwargs" in obj and len(obj) == 4

class JsonAst(BaseJsonAst, BaseDataTrans[dict]):

    def dumps(self) -> dict:
        return self.todict()
    
    @classmethod
    def loads(cls, obj: dict) -> "JsonAst":
        return cls.fromdict(obj)
